# clean.py
# ---------
# COSC4550/COSC5550
# Artificial Intelligence
# University of Wyoming
# ---------
# Most code was part of the Pacman AI projects by John DeNero and Dan Klein
# Code has been modified for the Cleaning Agents AI Challenge by Joost Huizinga
# Licensing Information: Please do not distribute or publish solutions to this
# project. You are free to use and extend these projects for educational
# purposes. The Pacman AI projects were developed at UC Berkeley, primarily by
# John DeNero (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# For more info, see http://inst.eecs.berkeley.edu/~cs188/sp09/pacman.html

"""
clean.py holds the logic for the Cleaning Agents AI Challeneg along with the main
code to run a game.  This file is divided into two sections:

  (i)  The game state and game rules:
          This section contains all of the logic code that the
          environment uses to decide who can move where, etc.  
          You shouldn't need to read this section
          of code, but you can if you want.

  (ii) Framework to start a game:
          The final section contains the code for reading the command
          you use to set up the game, then starting up a new game, along with
          linking in all the external parts (agent functions, graphics).
          Check this section out to see all the options available to you.
"""
from game import GameStateData
from game import Game
from game import Directions
from game import Actions
from game import RobotActions
from util import nearestPoint
from util import manhattanDistance
import util, layout
import sys, types, time, random, os, copy

###################################################
#            Game State and Game Rules            #
###################################################

class GameState:
  """
  GameStates are used by the Game object to capture the actual state of the game and
  can be used by agents to reason about the game.

  Much of the information in a GameState is stored in a GameStateData object.  We
  strongly suggest that you access that data via the accessor methods below rather
  than referring to the GameStateData object directly.
  """

  def getLegalActions( self, agentIndex=0 ):
    """
    Returns the legal actions for the agent specified.
    """
    if self.isFinished(): return []
    return RobotRules.getLegalActions( self, agentIndex )

  def generateSuccessor( self, agentIndex, action):
    """
    Returns the successor state after the specified agent takes the action.
    """
    # Check that successors exist
    if self.isFinished(): raise Exception('Can\'t generate a successor of a terminal state.')

    # Copy current state
    state = GameState(self)

    # Let agent's logic deal with its action's effects on the board
    RobotRules.applyAction( state, action, agentIndex)

    # Time passes
    state.data.timer += -1
    if(state.data.timer <=0):
      state.data._finished = True

    # Book keeping
    state.data._agentMoved = agentIndex
    state.data.score += state.data.scoreChange
    return state

  def moveDust(self, draft):
    state = GameState(self)
    state.data.newDustPos = []
    state.data.prevDustPos = []

    if not(draft == (0,0)):
      for i in range(len(self.data.dustPos)):
        dustPos = self.data.dustPos[i]
        (x, y) = dustPos
        (tempx, tempy) = dustPos
        (xmod, ymod) = draft

        while self.data.dust[tempx + xmod][tempy+ymod]:
          tempx += xmod
          tempy += ymod

        if(not(self.getWalls()[tempx + xmod][tempy+ymod])):
          state.data.prevDustPos.append((x, y))
          y += ymod
          x += xmod
          state.data.newDustPos.append((x, y))
          state.data.dustPos[i] = (x, y)

    for dustPos in state.data.prevDustPos:
      (x, y) = dustPos
      state.data.dust[x][y] = False

    for dustPos in state.data.newDustPos:
      (x, y) = dustPos
      state.data.dust[x][y] = True

    return state


  def getNumAgents( self ):
    return len( self.data.agentStates )

  def getScore( self ):
    return self.data.score

  def getNumDust( self ):
    return self.data.dust.count()

  def getDust(self):
    """
    Returns a Grid of boolean food indicator variables.

    Grids can be accessed via list notation, so to check
    if there is dust at (x,y), just call

    currentDust = state.getDust()
    if currentDust[x][y] == True: ...
    """
    return self.data.dust

  def getWalls(self):
    """
    Returns a Grid of boolean wall indicator variables.

    Grids can be accessed via list notation, so to check
    if there is food at (x,y), just call

    walls = state.getWalls()
    if walls[x][y] == True: ...
    """
    return self.data.layout.walls

  def hasDust(self, x, y):
    return self.data.dust[x][y]

  def hasWall(self, x, y):
    return self.data.layout.walls[x][y]

  def isFinished(self):
    return self.data._finished

  #############################################
  #             Helper methods:               #
  # You shouldn't need to call these directly #
  #############################################

  def __init__( self, prevState = None ):
    """
    Generates a new state by copying information from its predecessor.
    """
    if prevState != None: # Initial state
      self.data = GameStateData(prevState.data)
    else:
      self.data = GameStateData()

  def deepCopy( self ):
    state = GameState( self )
    state.data = self.data.deepCopy()
    return state

  def __eq__( self, other ):
    """
    Allows two states to be compared.
    """
    return self.data == other.data

  def __hash__( self ):
    """
    Allows states to be keys of dictionaries.
    """
    return hash( self.data )

  def __str__( self ):

    return str(self.data)

  def initialize( self, layout):
    """
    Creates an initial game state from a layout array (see layout.py).
    """
    self.data.initialize(layout)

TIME_PENALTY = 1 # Number of points lost each round

class CleanerGameRules:
  """
  These game rules manage the control flow of a game, deciding when
  and how the game starts and ends.
  """
  def __init__(self, timeout=30):
    self.timeout = timeout

  def newGame( self, layout, mainAgent, display, quiet = False, catchExceptions=False, useDraft = False, useUtility = False):
    agents = [mainAgent]
    initState = GameState()
    initState.initialize( layout )
    game = Game(agents, display, self, catchExceptions=catchExceptions, useDraft=useDraft)
    game.state = initState
    game.state.data.utility = useUtility
    self.initialState = initState.deepCopy()
    self.quiet = quiet
    return game

  def process(self, state, game):
    """
    Checks to see whether it is time to end the game.
    """
    if state.isFinished(): self.finish(state, game)

  def finish( self, state, game ):
    if not self.quiet: print "Trail is finished! Score: %f" % state.data.score
  #  if not self.quiet and state.data.utility: print "Utility for collecting dust: %d" % state.data.dustUtility


    game.gameOver = True

  def getProgress(self, game):
    return float(game.state.getNumDust()) / self.initialState.getNumDust()

  def agentCrash(self, game, agentIndex):
    print "An agent crashed"

  def getMaxTotalTime(self, agentIndex):
    return self.timeout

  def getMaxStartupTime(self, agentIndex):
    return self.timeout

  def getMoveWarningTime(self, agentIndex):
    return self.timeout

  def getMoveTimeout(self, agentIndex):
    return self.timeout

  def getMaxTimeWarnings(self, agentIndex):
    return 0







class RobotRules:
  """
  These functions govern how pacman interacts with his environment under
  the classic game rules.
  """
  SPEED=1


  def getLegalActions( state, index ):
    """
    Returns a list of possible actions.
    """
    return Actions.getPossibleActions( state.data.agentStates[index].configuration, state.data.layout.walls )
  getLegalActions = staticmethod( getLegalActions )

  def applyAction( state, action, index ):
    """
    Edits the state to reflect the results of the action.
    """
    agentState = state.data.agentStates[index]  

    # Clean
    next = agentState.configuration.getPosition()
    nearest = nearestPoint( next )
    if manhattanDistance( nearest, next ) <= 0.5 :
      RobotRules.consume( nearest, state )

   
    newDirection = agentState.configuration.direction
    newVector = Actions.directionToVector( Directions.STOP, RobotRules.SPEED )



    if action == RobotActions.TURN_LEFT or action == RobotActions.TURN_AND_MOVE_LEFT:
      newDirection = Directions.LEFT[agentState.configuration.direction]
    elif action == RobotActions.TURN_RIGHT or action == RobotActions.TURN_AND_MOVE_RIGHT:
      newDirection = Directions.RIGHT[agentState.configuration.direction]
    
    if (action == RobotActions.DO_NOTHING or action == RobotActions.STOP) and state.data.utility:
      state.data.scoreChange += float(state.data.energyUtility) / (float(state.data.dustUtility)*float(state.data.totalDust) + 200.0)

    if action == RobotActions.FORWARD or action == RobotActions.TURN_AND_MOVE_LEFT or action == RobotActions.TURN_AND_MOVE_RIGHT:
      legal = RobotRules.getLegalActions( state, index )
      if newDirection in legal:
        newVector = Actions.directionToVector( newDirection, RobotRules.SPEED )
      
    
    agentState.configuration = agentState.configuration.generateSuccessor(newDirection, newVector )

    # Clean
    next = agentState.configuration.getPosition()
    nearest = nearestPoint( next )
    if manhattanDistance( nearest, next ) <= 0.5 :
      RobotRules.consume( nearest, state )
  applyAction = staticmethod( applyAction )

  def consume( position, state ):
    x,y = position
    # Remove dust
    if state.data.dust[x][y]:

      if(state.data.utility):
        state.data.scoreChange += float(state.data.dustUtility) / (float(state.data.dustUtility)*float(state.data.totalDust) + 200.0)
      else:
        state.data.scoreChange += 1.0/float(state.data.totalDust)        

      state.data.dust[x][y] = False
      state.data.dustPos.remove((x,y))
      if state.data.newDustPos.count((x,y)):
        state.data.newDustPos.remove((x,y))
      state.data.prevDustPos.append((x,y))
      numDust = state.getNumDust()
      if numDust ==  0 and not state.data.utility:
        state.data._finished = True

  consume = staticmethod( consume )


#############################
# FRAMEWORK TO START A GAME #
#############################

def default(str):
  return str + ' [Default: %default]'

def parseAgentArgs(str):
  if str == None: return {}
  pieces = str.split(',')
  opts = {}
  for p in pieces:
    if '=' in p:
      key, val = p.split('=')
    else:
      key,val = p, 1
    opts[key] = val
  return opts

def readCommand( argv ):
  """
  Processes the command used to run pacman from the command line.
  """
  from optparse import OptionParser
  usageStr = """
  USAGE:      python clean.py <options>
  EXAMPLES:   (1) python clean.py
                  - starts a single run using the BasicAgent
              (2) python clean.py --agent SimpleReflexAgent
                  - starts a single run using the SimpleReflexAgent (that you have to implement)
              (3) python clean.py --agent ModelBasedReflexAgent -n 100 -q
                  - starts 100 silent runs using the ModelBasedReflexAgent (that you have to implement)
  """
  parser = OptionParser(usageStr)

  parser.add_option('-n', '--numGames', dest='numGames', type='int',
                    help=default('the number of GAMES to play'), metavar='GAMES', default=1)
  parser.add_option('-l', '--layout', dest='layout',
                    help=default('the LAYOUT_FILE from which to load the map layout'),
                    metavar='LAYOUT_FILE', default='mediumClassic')
  parser.add_option('-p', '--agent', dest='agent',
                    help=default('the agent TYPE to use'),
                    metavar='TYPE', default='BasicAgent')
  parser.add_option('-t', '--textGraphics', action='store_true', dest='textGraphics',
                    help='Display output as text only', default=False)
  parser.add_option('-q', '--quietTextGraphics', action='store_true', dest='quietGraphics',
                    help='Generate minimal output and no graphics', default=False)
  parser.add_option('-z', '--zoom', type='float', dest='zoom',
                    help=default('Zoom the size of the graphics window'), default=1.0)
  parser.add_option('-f', '--fixRandomSeed', action='store_true', dest='fixRandomSeed',
                    help='Fixes the random seed to always play the same game', default=False)
  parser.add_option('-r', '--recordActions', action='store_true', dest='record',
                    help='Writes game histories to a file (named by the time they were played)', default=False)
  parser.add_option('--replay', dest='gameToReplay',
                    help='A recorded game file (pickle) to replay', default=None)
  parser.add_option('-a','--agentArgs',dest='agentArgs',
                    help='Comma separated values sent to agent. e.g. "opt1=val1,opt2,opt3=val3"')
  parser.add_option('-x', '--numTraining', dest='numTraining', type='int',
                    help=default('How many episodes are training (suppresses output)'), default=0)
  parser.add_option('--frameTime', dest='frameTime', type='float',
                    help=default('Time to delay between frames; <0 means keyboard'), default=0.1)
  parser.add_option('-c', '--catchExceptions', action='store_true', dest='catchExceptions', 
                    help='Turns on exception handling and timeouts during games', default=False)
  parser.add_option('--timeout', dest='timeout', type='int',
                    help=default('Maximum length of time an agent can spend computing in a single game'), default=30)
  parser.add_option('--writeScore', action='store_true', dest='writeScore',
                    help=default('Write the final score to file for grading'), default=False)

  options, otherjunk = parser.parse_args(argv)
  if len(otherjunk) != 0:
    raise Exception('Command line input not understood: ' + str(otherjunk))
  args = dict()

  
  if options.writeScore:
    args['writeScore'] = True

  # Fix the random seed
  if options.fixRandomSeed: random.seed('cs188')

  # Choose a layout
  args['layoutOptions'] = options.layout
  if args['layoutOptions'] == None: raise Exception("The layout " + options.layout + " cannot be found")

  # Choose an agent
  if options.agent == 'DynamicWorldAgent':
    args['useDraft'] = True
  else:
    args['useDraft'] = False

  if options.agent == 'UtilityBasedAgent':
    args['useUtility'] = True
  else:
    args['useUtility'] = False
#  agent = loadAgent(options.agent)
  agentOpts = parseAgentArgs(options.agentArgs)

#  agent = agent(**agentOpts) # Instantiate agent with agentArgs
  args['agent'] = options.agent
  args['agentArgs'] = agentOpts

  if options.numTraining > 0:
    args['numTraining'] = options.numTraining
    if 'numTraining' not in agentOpts: agentOpts['numTraining'] = options.numTraining

  # Don't display training games
  if 'numTrain' in agentOpts:
    options.numQuiet = int(agentOpts['numTrain'])
    options.numIgnore = int(agentOpts['numTrain'])

  # Choose a display format
  if options.quietGraphics:
      import textDisplay
      args['display'] = textDisplay.NullGraphics()
  elif options.textGraphics:
    import textDisplay
    textDisplay.SLEEP_TIME = options.frameTime
    args['display'] = textDisplay.PacmanGraphics()
  else:
    import graphicsDisplay
    args['display'] = graphicsDisplay.AgentGraphics(options.zoom, frameTime = options.frameTime)
  args['numGames'] = options.numGames
  args['record'] = options.record
  args['catchExceptions'] = options.catchExceptions
  args['timeout'] = options.timeout

  # Special case: recorded games don't use the runGames method or args structure
  if options.gameToReplay != None:
    print 'Replaying recorded game %s.' % options.gameToReplay
    import cPickle
    f = open(options.gameToReplay)
    try: recorded = cPickle.load(f)
    finally: f.close()
    recorded['display'] = args['display']
    replayGame(**recorded)
    sys.exit(0)

  return args

def loadAgent(agent):
  # Looks through all pythonPath Directories for the right module,
  pythonPathStr = os.path.expandvars("$PYTHONPATH")
  if pythonPathStr.find(';') == -1:
    pythonPathDirs = pythonPathStr.split(':')
  else:
    pythonPathDirs = pythonPathStr.split(';')
  pythonPathDirs.append('.')


  for moduleDir in pythonPathDirs:
    if not os.path.isdir(moduleDir): continue
    moduleNames = [f for f in os.listdir(moduleDir) if f.endswith('gents.py')]
    for modulename in moduleNames:
      module = __import__(modulename[:-3])
      try:
        module = __import__(modulename[:-3])
      except ImportError:
        print "Error loading:"
        continue
      if agent in dir(module):
        return getattr(module, agent)
  raise Exception('The agent ' + agent + ' is not specified in any *Agents.py.')

def replayGame( layout, actions, display ):
    import pacmanAgents, ghostAgents
    rules = ClassicGameRules()
    agents = [pacmanAgents.GreedyAgent()] + [ghostAgents.RandomGhost(i+1) for i in range(layout.getNumGhosts())]
    game = rules.newGame( layout, agents[0], agents[1:], display )
    state = game.state
    display.initialize(state.data)

    for action in actions:
      # Execute the action
      state = state.generateSuccessor( *action )
      # Change the display
      display.update( state.data )
      # Allow for game specific conditions (winning, losing, etc.)
      rules.process(state, game)

    display.finish()

def runGames( layoutOptions, agent, agentArgs, display, numGames, record, numTraining = 0, catchExceptions=False, timeout=30, useDraft=False, useUtility = False, writeScore = False ):
  import __main__
  __main__.__dict__['_display'] = display

  rules = CleanerGameRules(timeout)
  games = []

  for i in range( numGames ):
    beQuiet = i < numTraining
    if beQuiet:
        # Suppress output and graphics
        import textDisplay
        gameDisplay = textDisplay.NullGraphics()
        rules.quiet = True
    else:
        gameDisplay = display
        rules.quiet = False
    newLayout = layout.getLayout( layoutOptions )

    newAgent = loadAgent(agent)
    newAgent = newAgent(**agentArgs)

    RobotRules.dustUtility = 4
    game = rules.newGame( newLayout, newAgent, gameDisplay, beQuiet, catchExceptions, useDraft, useUtility)
    game.run()
    if not beQuiet: games.append(game)

    if record:
      import time, cPickle
      fname = ('recorded-game-%d' % (i + 1)) +  '-'.join([str(t) for t in time.localtime()[1:6]])
      f = file(fname, 'w')
      components = {'layout': layout, 'actions': game.moveHistory}
      cPickle.dump(components, f)
      f.close()

  if numGames > 1:
    scores = [game.state.getScore() for game in games]
    print 'Average Score:', sum(scores) / float(len(scores))
    print 'Scores:       ', ', '.join([str(score) for score in scores])
    if writeScore:
      with open("score", 'a') as score_file:
        score_file.write(str(sum(scores) / float(len(scores))) + " ")

  return games

if __name__ == '__main__':
  """
  The main function called when clean.py is run
  from the command line:

  > python clean.py

  See the usage string for more details.

  > python clean.py --help
  """
  args = readCommand( sys.argv[1:] ) # Get game components based on input
  runGames( **args )

  pass
